@@ -15,25 +15,29 @@ import sys
 import stat
 import errno
 import socket
+import gzip
+import math
 try:
     from hashlib import sha1
 except ImportError:
     from sha import new as sha1
 import traceback
+import collections
 from datetime import datetime, timedelta
 from collections import deque
 from textwrap import dedent
 
 from logbook.base import (
-    CRITICAL, ERROR, WARNING, NOTICE, INFO, DEBUG, NOTSET, level_name_property,
-    _missing, lookup_level, Flags, ContextObject, ContextStackManager)
+    CRITICAL, ERROR, WARNING, NOTICE, INFO, DEBUG, TRACE, NOTSET, level_name_property,
+    _missing, lookup_level, Flags, ContextObject, ContextStackManager,
+    _datetime_factory)
 from logbook.helpers import (
     rename, b, _is_text_stream, is_unicode, PY2, zip, xrange, string_types,
     integer_types, reraise, u, with_metaclass)
 from logbook.concurrency import new_fine_grained_lock
 
 DEFAULT_FORMAT_STRING = u(
-    '[{record.time:%Y-%m-%d %H:%M:%S.%f}] '
+    '[{record.time:%Y-%m-%d %H:%M:%S.%f%z}] '
     '{record.level_name}: {record.channel}: {record.message}')
 
 SYSLOG_FORMAT_STRING = u('{record.channel}: {record.message}')
@@ -494,8 +498,7 @@ class LimitingHandlerMixin(HashingHandlerMixin):
 
             if (not suppression_count and
                     len(self._record_limits) >= self.max_record_cache):
-                cache_items = self._record_limits.items()
-                cache_items.sort()
+                cache_items = sorted(self._record_limits.items())
                 del cache_items[:int(self._record_limits)
                                 * self.record_cache_prune]
                 self._record_limits = dict(cache_items)
@@ -581,10 +584,14 @@ class StreamHandler(Handler, StringFormatterHandlerMixin):
         try:
             self.ensure_stream_is_open()
             self.write(self.encode(msg))
-            self.flush()
+            if self.should_flush():
+                self.flush()
         finally:
             self.lock.release()
 
+    def should_flush(self):
+        return True
+
 
 class FileHandler(StreamHandler):
     """A handler that does the task of opening and closing files for you.
@@ -641,6 +648,79 @@ class FileHandler(StreamHandler):
             self._open()
 
 
+class GZIPCompressionHandler(FileHandler):
+    def __init__(self, filename, encoding=None, level=NOTSET,
+                 format_string=None, delay=False, filter=None, bubble=False, compression_quality=9):
+
+        self._compression_quality = compression_quality
+        super(GZIPCompressionHandler, self).__init__(filename, mode='wb', encoding=encoding, level=level,
+                             format_string=format_string, delay=delay, filter=filter, bubble=bubble)
+
+    def _open(self, mode=None):
+        if mode is None:
+            mode = self._mode
+        self.stream = gzip.open(self._filename, mode, compresslevel=self._compression_quality)
+
+    def write(self, item):
+        if isinstance(item, str):
+            item = item.encode(encoding=self.encoding)
+        self.ensure_stream_is_open()
+        self.stream.write(item)
+
+    def should_flush(self):
+        # gzip manages writes independently. Flushing prematurely could mean
+        # duplicate flushes and thus bloated files
+        return False
+
+
+class BrotliCompressionHandler(FileHandler):
+    def __init__(self, filename, encoding=None, level=NOTSET,
+                 format_string=None, delay=False, filter=None, bubble=False,
+                 compression_window_size=4*1024**2, compression_quality=11):
+        super(BrotliCompressionHandler, self).__init__(filename, mode='wb', encoding=encoding, level=level,
+                             format_string=format_string, delay=delay, filter=filter, bubble=bubble)
+        try:
+            from brotli import Compressor
+        except ImportError:
+            raise RuntimeError('The brotli library is required for '
+                               'the BrotliCompressionHandler.')
+
+        max_window_size = int(math.log(compression_window_size, 2))
+        self._compressor = Compressor(quality=compression_quality, lgwin=max_window_size)
+
+    def _open(self, mode=None):
+        if mode is None:
+            mode = self._mode
+        self.stream = io.open(self._filename, mode)
+
+    def write(self, item):
+        if isinstance(item, str):
+            item = item.encode(encoding=self.encoding)
+        ret = self._compressor.process(item)
+        if ret:
+            self.ensure_stream_is_open()
+            self.stream.write(ret)
+            super(BrotliCompressionHandler, self).flush()
+
+    def should_flush(self):
+        return False
+
+    def flush(self):
+        if self._compressor is not None:
+            ret = self._compressor.flush()
+            if ret:
+                self.ensure_stream_is_open()
+                self.stream.write(ret)
+        super(BrotliCompressionHandler, self).flush()
+
+    def close(self):
+        if self._compressor is not None:
+            self.ensure_stream_is_open()
+            self.stream.write(self._compressor.finish())
+            self._compressor = None
+        super(BrotliCompressionHandler, self).close()
+
+
 class MonitoringFileHandler(FileHandler):
     """A file handler that will check if the file was moved while it was
     open.  This might happen on POSIX systems if an application like
@@ -667,7 +747,7 @@ class MonitoringFileHandler(FileHandler):
                 st = os.stat(self._filename)
             except OSError:
                 e = sys.exc_info()[1]
-                if e.errno != 2:
+                if e.errno != errno.ENOENT:
                     raise
                 self._last_stat = None, None
             else:
@@ -782,29 +862,72 @@ class TimedRotatingFileHandler(FileHandler):
 
     By default it will keep all these files around, if you want to limit
     them, you can specify a `backup_count`.
+
+    You may supply an optional `rollover_format`. This allows you to specify
+    the format for the filenames of rolled-over files.
+    the format as
+
+    So for example if you configure your handler like this::
+
+        handler = TimedRotatingFileHandler(
+            '/var/log/foo.log',
+            date_format='%Y-%m-%d',
+            rollover_format='{basename}{ext}.{timestamp}')
+
+    The filenames for the logfiles will look like this::
+
+        /var/log/foo.log.2010-01-10
+        /var/log/foo.log.2010-01-11
+        ...
+
+    Finally, an optional argument `timed_filename_for_current` may be set to
+    false if you wish to have the current log file match the supplied filename
+    until it is rolled over
     """
 
     def __init__(self, filename, mode='a', encoding='utf-8', level=NOTSET,
                  format_string=None, date_format='%Y-%m-%d',
-                 backup_count=0, filter=None, bubble=False):
-        FileHandler.__init__(self, filename, mode, encoding, level,
-                             format_string, True, filter, bubble)
+                 backup_count=0, filter=None, bubble=False,
+                 timed_filename_for_current=True,
+                 rollover_format='{basename}-{timestamp}{ext}'):
         self.date_format = date_format
         self.backup_count = backup_count
-        self._fn_parts = os.path.splitext(os.path.abspath(filename))
-        self._filename = None
 
-    def _get_timed_filename(self, datetime):
-        return (datetime.strftime('-' + self.date_format)
-                .join(self._fn_parts))
+        self.rollover_format = rollover_format
 
-    def should_rollover(self, record):
-        fn = self._get_timed_filename(record.time)
-        rv = self._filename is not None and self._filename != fn
-        # remember the current filename.  In case rv is True, the rollover
-        # performing function will already have the new filename
-        self._filename = fn
-        return rv
+        self.original_filename = filename
+        self.basename, self.ext = os.path.splitext(os.path.abspath(filename))
+        self.timed_filename_for_current = timed_filename_for_current
+
+        self._timestamp = self._get_timestamp(_datetime_factory())
+        if self.timed_filename_for_current:
+            filename = self.generate_timed_filename(self._timestamp)
+        elif os.path.exists(filename):
+            self._timestamp = self._get_timestamp(
+                datetime.fromtimestamp(
+                    os.stat(filename).st_mtime
+                )
+            )
+
+        FileHandler.__init__(self, filename, mode, encoding, level,
+                             format_string, True, filter, bubble)
+
+    def _get_timestamp(self, datetime):
+        """
+        Fetches a formatted string witha timestamp of the given datetime
+        """
+        return datetime.strftime(self.date_format)
+
+    def generate_timed_filename(self, timestamp):
+        """
+        Produces a filename that includes a timestamp in the format supplied
+        to the handler at init time.
+        """
+        timed_filename = self.rollover_format.format(
+            basename=self.basename,
+            timestamp=timestamp,
+            ext=self.ext)
+        return timed_filename
 
     def files_to_delete(self):
         """Returns a list with the files that have to be deleted when
@@ -812,10 +935,14 @@ class TimedRotatingFileHandler(FileHandler):
         """
         directory = os.path.dirname(self._filename)
         files = []
+        rollover_regex = re.compile(self.rollover_format.format(
+            basename=re.escape(self.basename),
+            timestamp='.+',
+            ext=re.escape(self.ext),
+        ))
         for filename in os.listdir(directory):
             filename = os.path.join(directory, filename)
-            if (filename.startswith(self._fn_parts[0] + '-') and
-                    filename.endswith(self._fn_parts[1])):
+            if rollover_regex.match(filename):
                 files.append((os.path.getmtime(filename), filename))
         files.sort()
         if self.backup_count > 1:
@@ -823,19 +950,34 @@ class TimedRotatingFileHandler(FileHandler):
         else:
             return files[:]
 
-    def perform_rollover(self):
-        self.stream.close()
+    def perform_rollover(self, new_timestamp):
+        if self.stream is not None:
+            self.stream.close()
+
+        if (
+                not self.timed_filename_for_current
+                and os.path.exists(self._filename)
+        ):
+            filename = self.generate_timed_filename(self._timestamp)
+            os.rename(self._filename, filename)
+
         if self.backup_count > 0:
             for time, filename in self.files_to_delete():
                 os.remove(filename)
+
+        if self.timed_filename_for_current:
+            self._filename = self.generate_timed_filename(new_timestamp)
+        self._timestamp = new_timestamp
+
         self._open('w')
 
     def emit(self, record):
         msg = self.format(record)
         self.lock.acquire()
         try:
-            if self.should_rollover(record):
-                self.perform_rollover()
+            new_timestamp = self._get_timestamp(record.time)
+            if new_timestamp != self._timestamp:
+                self.perform_rollover(new_timestamp)
             self.write(self.encode(msg))
             self.flush()
         finally:
@@ -920,6 +1062,11 @@ class TestHandler(Handler, StringFormatterHandlerMixin):
         """`True` if any :data:`DEBUG` records were found."""
         return any(r.level == DEBUG for r in self.records)
 
+    @property
+    def has_traces(self):
+        """`True` if any :data:`TRACE` records were found."""
+        return any(r.level == TRACE for r in self.records)
+
     def has_critical(self, *args, **kwargs):
         """`True` if a specific :data:`CRITICAL` log record exists.
 
@@ -968,9 +1115,17 @@ class TestHandler(Handler, StringFormatterHandlerMixin):
         kwargs['level'] = DEBUG
         return self._test_for(*args, **kwargs)
 
+    def has_trace(self, *args, **kwargs):
+        """`True` if a specific :data:`TRACE` log record exists.
+
+        See :ref:`probe-log-records` for more information.
+        """
+        kwargs['level'] = TRACE
+        return self._test_for(*args, **kwargs)
+
     def _test_for(self, message=None, channel=None, level=None):
         def _match(needle, haystack):
-            "Matches both compiled regular expressions and strings"
+            """Matches both compiled regular expressions and strings"""
             if isinstance(needle, REGTYPE) and needle.search(haystack):
                 return True
             if needle == haystack:
@@ -1014,14 +1169,42 @@ class MailHandler(Handler, StringFormatterHandlerMixin,
 
     The default timedelta is 60 seconds (one minute).
 
-    The mail handler is sending mails in a blocking manner.  If you are not
+    The mail handler sends mails in a blocking manner.  If you are not
     using some centralized system for logging these messages (with the help
     of ZeroMQ or others) and the logging system slows you down you can
     wrap the handler in a :class:`logbook.queues.ThreadedWrapperHandler`
     that will then send the mails in a background thread.
 
+    `server_addr` can be a tuple of host and port, or just a string containing
+    the host to use the default port (25, or 465 if connecting securely.)
+
+    `credentials` can be a tuple or dictionary of arguments that will be passed
+    to :py:meth:`smtplib.SMTP.login`.
+
+    `secure` can be a tuple, dictionary, or boolean. As a boolean, this will
+    simply enable or disable a secure connection. The tuple is unpacked as
+    parameters `keyfile`, `certfile`. As a dictionary, `secure` should contain
+    those keys. For backwards compatibility, ``secure=()`` will enable a secure
+    connection. If `starttls` is enabled (default), these parameters will be
+    passed to :py:meth:`smtplib.SMTP.starttls`, otherwise
+    :py:class:`smtplib.SMTP_SSL`.
+
+
     .. versionchanged:: 0.3
        The handler supports the batching system now.
+
+    .. versionadded:: 1.0
+       `starttls` parameter added to allow disabling STARTTLS for SSL
+       connections.
+
+    .. versionchanged:: 1.0
+       If `server_addr` is a string, the default port will be used.
+
+    .. versionchanged:: 1.0
+       `credentials` parameter can now be a dictionary of keyword arguments.
+
+    .. versionchanged:: 1.0
+       `secure` can now be a dictionary or boolean in addition to to a tuple.
     """
     default_format_string = MAIL_FORMAT_STRING
     default_related_format_string = MAIL_RELATED_FORMAT_STRING
@@ -1039,7 +1222,7 @@ class MailHandler(Handler, StringFormatterHandlerMixin,
                  server_addr=None, credentials=None, secure=None,
                  record_limit=None, record_delta=None, level=NOTSET,
                  format_string=None, related_format_string=None,
-                 filter=None, bubble=False):
+                 filter=None, bubble=False, starttls=True):
         Handler.__init__(self, level, filter, bubble)
         StringFormatterHandlerMixin.__init__(self, format_string)
         LimitingHandlerMixin.__init__(self, record_limit, record_delta)
@@ -1054,6 +1237,7 @@ class MailHandler(Handler, StringFormatterHandlerMixin,
         if related_format_string is None:
             related_format_string = self.default_related_format_string
         self.related_format_string = related_format_string
+        self.starttls = starttls
 
     def _get_related_format_string(self):
         if isinstance(self.related_formatter, StringFormatter):
@@ -1141,27 +1325,70 @@ class MailHandler(Handler, StringFormatterHandlerMixin,
             mail.get_payload(),
             title,
             '\r\n\r\n'.join(body.rstrip() for body in related)
-        ))
+        ), 'UTF-8')
         return mail
 
     def get_connection(self):
         """Returns an SMTP connection.  By default it reconnects for
         each sent mail.
         """
-        from smtplib import SMTP, SMTP_PORT, SMTP_SSL_PORT
+        from smtplib import SMTP, SMTP_SSL, SMTP_PORT, SMTP_SSL_PORT
         if self.server_addr is None:
             host = '127.0.0.1'
             port = self.secure and SMTP_SSL_PORT or SMTP_PORT
         else:
-            host, port = self.server_addr
-        con = SMTP()
-        con.connect(host, port)
+            try:
+                host, port = self.server_addr
+            except ValueError:
+                # If server_addr is a string, the tuple unpacking will raise
+                # ValueError, and we can use the default port.
+                host = self.server_addr
+                port = self.secure and SMTP_SSL_PORT or SMTP_PORT
+
+        # Previously, self.secure was passed as con.starttls(*self.secure). This
+        # meant that starttls couldn't be used without a keyfile and certfile
+        # unless an empty tuple was passed. See issue #94.
+        #
+        # The changes below allow passing:
+        # - secure=True for secure connection without checking identity.
+        # - dictionary with keys 'keyfile' and 'certfile'.
+        # - tuple to be unpacked to variables keyfile and certfile.
+        # - secure=() equivalent to secure=True for backwards compatibility.
+        # - secure=False equivalent to secure=None to disable.
+        if isinstance(self.secure, collections.Mapping):
+            keyfile = self.secure.get('keyfile', None)
+            certfile = self.secure.get('certfile', None)
+        elif isinstance(self.secure, collections.Iterable):
+            # Allow empty tuple for backwards compatibility
+            if len(self.secure) == 0:
+                keyfile = certfile = None
+            else:
+                keyfile, certfile = self.secure
+        else:
+            keyfile = certfile = None
+
+        # Allow starttls to be disabled by passing starttls=False.
+        if not self.starttls and self.secure:
+            con = SMTP_SSL(host, port, keyfile=keyfile, certfile=certfile)
+        else:
+            con = SMTP(host, port)
+
         if self.credentials is not None:
-            if self.secure is not None:
+            secure = self.secure
+            if self.starttls and secure is not None and secure is not False:
                 con.ehlo()
-                con.starttls(*self.secure)
+                con.starttls(keyfile=keyfile, certfile=certfile)
                 con.ehlo()
-            con.login(*self.credentials)
+
+            # Allow credentials to be a tuple or dict.
+            if isinstance(self.credentials, collections.Mapping):
+                credentials_args = ()
+                credentials_kwargs = self.credentials
+            else:
+                credentials_args = self.credentials
+                credentials_kwargs = dict()
+
+            con.login(*credentials_args, **credentials_kwargs)
         return con
 
     def close_connection(self, con):
@@ -1175,7 +1402,7 @@ class MailHandler(Handler, StringFormatterHandlerMixin,
             pass
 
     def deliver(self, msg, recipients):
-        """Delivers the given message to a list of recpients."""
+        """Delivers the given message to a list of recipients."""
         con = self.get_connection()
         try:
             con.sendmail(self.from_addr, recipients, msg.as_string())
@@ -1227,7 +1454,7 @@ class GMailHandler(MailHandler):
 
     def __init__(self, account_id, password, recipients, **kw):
         super(GMailHandler, self).__init__(
-            account_id, recipients, secure=(),
+            account_id, recipients, secure=True,
             server_addr=("smtp.gmail.com", 587),
             credentials=(account_id, password), **kw)
 
@@ -1307,7 +1534,7 @@ class SyslogHandler(Handler, StringFormatterHandlerMixin):
     def __init__(self, application_name=None, address=None,
                  facility='user', socktype=socket.SOCK_DGRAM,
                  level=NOTSET, format_string=None, filter=None,
-                 bubble=False):
+                 bubble=False, record_delimiter=None):
         Handler.__init__(self, level, filter, bubble)
         StringFormatterHandlerMixin.__init__(self, format_string)
         self.application_name = application_name
@@ -1318,14 +1545,24 @@ class SyslogHandler(Handler, StringFormatterHandlerMixin):
             else:
                 address = '/dev/log'
 
-        self.address = address
+        self.remote_address = self.address = address
         self.facility = facility
         self.socktype = socktype
 
         if isinstance(address, string_types):
             self._connect_unixsocket()
+            self.enveloper = self.unix_envelope
+            default_delimiter = u'\x00'
         else:
             self._connect_netsocket()
+            self.enveloper = self.net_envelope
+            default_delimiter = u'\n'
+
+        self.record_delimiter = default_delimiter \
+            if record_delimiter is None else record_delimiter
+
+        self.connection_exception = getattr(
+            __builtins__, 'BrokenPipeError', socket.error)
 
     def _connect_unixsocket(self):
         self.unixsocket = True
@@ -1341,7 +1578,7 @@ class SyslogHandler(Handler, StringFormatterHandlerMixin):
         self.unixsocket = False
         self.socket = socket.socket(socket.AF_INET, self.socktype)
         if self.socktype == socket.SOCK_STREAM:
-            self.socket.connect(self.address)
+            self.socket.connect(self.remote_address)
             self.address = self.socket.getsockname()
 
     def encode_priority(self, record):
@@ -1350,15 +1587,44 @@ class SyslogHandler(Handler, StringFormatterHandlerMixin):
                                                self.LOG_WARNING)
         return (facility << 3) | priority
 
-    def emit(self, record):
-        prefix = u('')
-        if self.application_name is not None:
-            prefix = self.application_name + u(':')
-        self.send_to_socket((u('<%d>%s%s\x00') % (
+    def wrap_segments(self, record, before):
+        msg = self.format(record)
+        segments = [segment for segment in msg.split(self.record_delimiter)]
+        return (before + segment + self.record_delimiter
+                for segment in segments)
+        
+    def unix_envelope(self, record):
+        before = u'<{}>{}'.format(
             self.encode_priority(record),
-            prefix,
-            self.format(record)
-        )).encode('utf-8'))
+            self.application_name + ':' if self.application_name else '')
+        return self.wrap_segments(record, before)
+
+    def net_envelope(self, record):
+        # Gross but effective
+        try:
+            format_string = self.format_string
+            application_name = self.application_name
+            if not application_name and record.channel and \
+               '{record.channel}: ' in format_string:
+                self.format_string = format_string.replace(
+                    '{record.channel}: ', '')
+                self.application_name = record.channel
+            # RFC 5424: <PRIVAL>version timestamp hostname app-name procid
+            #           msgid structured-data message
+            before = u'<{}>1 {}Z {} {} {} - - '.format(
+                self.encode_priority(record),
+                record.time.isoformat(),
+                socket.gethostname(),
+                self.application_name if self.application_name else '-',
+                record.process)
+            return self.wrap_segments(record, before)
+        finally:
+            self.format_string = format_string
+            self.application_name = application_name
+
+    def emit(self, record):
+        for segment in self.enveloper(record):
+            self.send_to_socket(segment.encode('utf-8'))
 
     def send_to_socket(self, data):
         if self.unixsocket:
@@ -1371,7 +1637,11 @@ class SyslogHandler(Handler, StringFormatterHandlerMixin):
             # the flags are no longer optional on Python 3
             self.socket.sendto(data, 0, self.address)
         else:
-            self.socket.sendall(data)
+            try:
+                self.socket.sendall(data)
+            except self.connection_exception:
+                self._connect_netsocket()
+                self.socket.send(data)
 
     def close(self):
         self.socket.close()
@@ -1431,9 +1701,15 @@ class NTEventLogHandler(Handler, StringFormatterHandlerMixin):
         return self._type_map.get(record.level, self._default_type)
 
     def get_event_category(self, record):
+        """Returns the event category for the record. Override this if you want
+        to specify your own categories. This version returns 0.
+        """
         return 0
 
     def get_message_id(self, record):
+        """Returns the message ID (EventID) for the record. Override this if
+        you want to specify your own ID. This version returns 1.
+        """
         return 1
 
     def emit(self, record):
